/*
##############################################################################
# MIT License
#
# Copyright (c) 2025 Advanced Micro Devices, Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

##############################################################################
*/
#pragma once

#include <hip/hip_runtime.h>
#include <iostream>

#define hipCheck(stmt)                                                         \
  do {                                                                         \
    hipError_t err = stmt;                                                     \
    if (err != hipSuccess) {                                                   \
      char msg[256];                                                           \
      sprintf(msg, "%s in file %s, function %s, line %d\n", #stmt, __FILE__,   \
              __FUNCTION__, __LINE__);                                         \
      std::string errstring = hipGetErrorString(err);                          \
      std::cerr << msg << "\t" << errstring << std::endl;                      \
      throw std::runtime_error(msg);                                           \
    }                                                                          \
  } while (0)
